/*
 *
 * btdev-voice.c
 *
 *
 *
 * Copyright (C) 2013 Broadcom Corporation.
 *
 *
 *
 * This software is licensed under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation (the "GPL"), and may
 * be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GPL for more details.
 *
 *
 * A copy of the GPL is available at http://www.broadcom.com/licenses/GPLv2.php
 * or by writing to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA
 *
 *
 */

// for kmalloc
#include <linux/slab.h>
#include "btusb.h"
#include "hcidefs.h"

/*******************************************************************************
 **
 ** Function         btusb_enqueue_voice
 **
 ** Description      Add a received voice message into the user queue
 **
 ** Parameters       p_dev: device instance control block
 **                  buffer: received buffer content
 **                  len: length of the content
 **
 ** Returns          void
 **
 *******************************************************************************/
static void btusb_enqueue_voice(tBTUSB_CB *p_dev, void *buffer, USHORT len)
{
    UINT8 *p = buffer;
    UINT16 handle;
    UINT8 sco_len;
    UINT8 packet_len = len;
    BT_HDR *p_buf = NULL;
    unsigned char *p1;

    BTUSB_INFO("len = %d\n", p_dev->RxCount);

    // we will save 15ms of SCO data before sending to app
    // so we do not have to interrupt the userial interface every
    // 3 ms....  This is 240 bytes of sco data + 3 for sco header
    if (packet_len == 0)
        return;

    if (p_dev->RxCount == 0)
    {
        // copy in sco handle + sco len + data to start of buffer
        memcpy(&p_dev->RxBuffer[p_dev->RxCount], p, packet_len);
        p_dev->RxCount = len;
    }
    else
    {
        // extract and skip the HCI handle and sco_length
        STREAM_TO_UINT16(handle, p);
        STREAM_TO_UINT8(sco_len, p);

        packet_len -= 3;
        memcpy(&p_dev->RxBuffer[p_dev->RxCount], p, packet_len);
        p_dev->RxCount += packet_len;

        if (p_dev->RxCount >= (SCO_RX_MAX_LEN + 3))
        {
            // do not have space for new packet so send what we have to app for processing
            if ((p_buf = (BT_HDR *) GKI_getpoolbuf(HCI_SCO_POOL_ID)) != NULL)
            {
                p_buf->event = BT_EVT_TO_BTU_HCI_SCO;
                p_buf->len = p_dev->RxCount + 1; // sizeof(data_type + handle + len + data)
                p_buf->offset = 0;
                p_buf->layer_specific = 0;

                // add data type
                p1 = (char *) (p_buf + 1);
                *p1++ = HCIT_TYPE_SCO_DATA;
                //UINT8_TO_STREAM(p1, HCIT_TYPE_SCO_DATA);

                // modify data len
                p_dev->RxBuffer[2] = p_dev->RxCount - 3;

                // add sco packet to p_buf
                memcpy(p1, p_dev->RxBuffer, p_dev->RxCount);

                GKI_enqueue(&p_dev->rx_queue, p_buf);

                // notify RX event(in case of select/poll
                wake_up_interruptible(&p_dev->rx_wait_q);
            }

            p_dev->RxCount = 0;
        }
        else
        {
            BTUSB_INFO("RxCount not large enough %d\n", p_dev->RxCount);
            //memcpy(&p_dev->RxBuffer[p_dev->RxCount], p, packet_len);
            //p_dev->RxCount += packet_len;
        }

    }

    return;
}

/*******************************************************************************
 **
 ** Function         btusb_voice_hdr_match
 **
 ** Description      Useful routine to check 3 bytes against voice headers configured
 **                  Broadcom 2035 does generate up to twice as much data as it's been
 **                  configured so we're checking length against the 2 times what's
 **                  configured rather than just once.  Should eventually be fixed.
 **
 ** Parameters
 **
 ** Returns          TRUE upon success, FALSE else
 **
 *******************************************************************************/
static int btusb_voice_hdr_match(tBTUSB_CB *p_dev, unsigned char *p_burst)
{
    int idx;
    unsigned short sco_handle;
    unsigned char burst;
    tBTUSB_VOICE_CHANNEL *p_chan;

    // can't have zero data length in the header
    if (unlikely(p_burst[2] == 0))
        return FALSE;

    STREAM_TO_UINT16(sco_handle, p_burst);
    sco_handle &= 0x0fff;

    STREAM_TO_UINT8(burst, p_burst);

    for (idx = 0; idx < ARRAY_SIZE(p_dev->voice_channels); idx++)
    {
        p_chan = &p_dev->voice_channels[idx];
        if ((p_chan->used) &&
            (sco_handle == p_chan->info.sco_handle) &&
            (burst <= (2 * p_chan->info.burst)))
        {
            return TRUE;
        }
    }
    return FALSE;
}

/*******************************************************************************
 **
 ** Function         btusb_voicerx_complete
 **
 ** Description      Voice read (iso pipe) completion routine.
 **
 ** Parameters
 **
 ** Returns          void
 **
 *******************************************************************************/
void btusb_voicerx_complete(struct urb *p_urb)
{
    tBTUSB_TRANSACTION *p_trans = p_urb->context;
    tBTUSB_CB *p_dev = p_trans->context;
    int length;
    int i;
    unsigned char *p_packet, *p_frame;
    int pending_data_forwarded = FALSE;
    unsigned long flags;
    static int read_voice_active = FALSE;
    int packet_size;
    struct usb_iso_packet_descriptor *p_uipd;

    BTUSB_INFO("enter");
    read_lock_irqsave(&p_dev->xferlock, flags);

    btusb_voice_stats(&(p_dev->stats.voice_max_rx_rdy_delta_time), &(p_dev->stats.voice_min_rx_rdy_delta_time),
            &(p_dev->stats.voice_rx_rdy_delta_time), &(p_dev->stats.voice_last_rx_rdy_ts));

    p_dev->stats.voicerx_completed++;

    if (unlikely(!p_dev->p_main_intf || !p_dev->p_voice_in))
    {
        BTUSB_DBG("intf is down\n");
        goto unlock;
    }

    packet_size = le16_to_cpu(p_dev->p_voice_in->desc.wMaxPacketSize);

    if (unlikely(p_urb->status))
    {
        BTUSB_ERR("failure %d\n", p_urb->status);
        p_dev->stats.voicerx_completed_error++;
        goto unlock;
    }

    if (unlikely(read_voice_active))
    {
        BTUSB_ERR("PARALLEL!\n");
        p_dev->stats.voicerx_parallel_dpcs++;
    }
    else
        read_voice_active = TRUE;

    // total length
    for (i = 0, length = 0; i < p_urb->number_of_packets; i++)
    {
        p_uipd = &p_urb->iso_frame_desc[i];
        if (!p_uipd->status)
            length += p_uipd->actual_length;
    }

    if (unlikely(p_dev->scosniff_active))
    {
        struct btusb_scosniff *bs;

        bs = kmalloc(sizeof(struct btusb_scosniff) +
                (p_urb->number_of_packets * sizeof(p_urb->iso_frame_desc[0])) +
                p_urb->transfer_buffer_length, GFP_ATOMIC);
        if (bs)
        {
            bs->s = p_urb->start_frame;
            bs->n = p_urb->number_of_packets;
            bs->l = p_urb->transfer_buffer_length;
            // copy the descriptors
            memcpy(bs->d, p_urb->iso_frame_desc, bs->n * sizeof(p_urb->iso_frame_desc[0]));
            // then copy the content of the buffer
            memcpy(&bs->d[bs->n], p_urb->transfer_buffer, bs->l);
            list_add_tail(&bs->lh, &p_dev->scosniff_list);
            complete(&p_dev->scosniff_completion);
        }
        else
        {
            BTUSB_ERR("Failed allocating scosniff buffer");
        }
    }

    // analyze what we've got (and concatenate data)
    for (i = 0, length = 0, p_packet = p_urb->transfer_buffer; i < p_urb->number_of_packets; i++)
    {
        p_uipd = &p_urb->iso_frame_desc[i];
        if (p_uipd->status)
        {
            p_dev->stats.voicerx_bad_frames++;
            BTUSB_DBG("(%d) err %d offset %d len %d\n",
                    i, p_uipd->status, p_uipd->offset,
                    p_uipd->actual_length);
        }
        else
        {
            p_frame = p_urb->transfer_buffer + p_uipd->offset;
            if (p_frame != p_packet)
            {
                p_dev->stats.voicerx_moved_frames++;
                memmove(p_packet, p_frame, p_uipd->actual_length);
            }
            p_packet += p_uipd->actual_length;
            length += p_uipd->actual_length;
        }
    }

    // nothing to do if no data
    if (length <= BTUSB_VOICE_HEADER_SIZE)
    {
        BTUSB_DBG("no data reported, length %d VH %d\n", length, BTUSB_VOICE_HEADER_SIZE);
        read_voice_active = FALSE;

        // submit the next read to the same place and get out
        btusb_submit_voice_rx(p_dev, p_trans, GFP_ATOMIC);

        goto unlock;
    }

    p_dev->stats.voicerx_bytes += length;
    p_packet = p_urb->transfer_buffer;

    // find the beginning of the aligned data
    i = 0;

    // if something is pending, start looking beyond the expected remainder
    if (p_dev->voice_pending_size)
    {
        // rhat's Ok if it's a huge number in case of junk data -
        // we'll just not find anything in the loop below...
        i = p_dev->voice_pending_buffer[2] + BTUSB_VOICE_HEADER_SIZE - p_dev->voice_pending_size;
    }

    for (; i < (unsigned long) (length - BTUSB_VOICE_HEADER_SIZE); i++)
    {
        if (btusb_voice_hdr_match(p_dev, &p_packet[i]))
        {
            unsigned char *p_burst = &p_packet[i];

            // some additional sanity checks

#if 0
            // this is the very beginning of the first voice packet
            // it must fit in the USB packet received completely
            if ((i + VOICE_HEADER_SIZE + p_burst[2]) > (unsigned long) length)
            continue;
#endif

            // if there is room remaining in this packet after the
            // voice packet, it's got to be the beginning of the
            // next voice packet...
            if ((i + BTUSB_VOICE_HEADER_SIZE + p_burst[2] + BTUSB_VOICE_HEADER_SIZE) <= (unsigned long) length)
            {
                if (btusb_voice_hdr_match(p_dev, &p_packet[i + BTUSB_VOICE_HEADER_SIZE + p_burst[2]]))
                    goto forward_data;
                else
                    continue;
            }

            goto forward_data;
        }
    }

    // have not found any headers anywhere - highly unlikely!
    p_dev->voice_pending_size = 0;
    BTUSB_ERR("no voice headers found!\n");
    p_dev->stats.voicerx_disc_nohdr++;
    read_voice_active = FALSE;

    // submit the next read to the same place and get out
    btusb_submit_voice_rx(p_dev, p_trans, GFP_ATOMIC);

    goto unlock;

forward_data:
    // are we in sync and got something large enough?
    if ((i == 0) && ((length - i) >= p_packet[i + 2]))
    {
        unsigned char *p_burst = p_packet;

        while (1)
        {
            if (btusb_voice_hdr_match(p_dev, p_burst))
            {
                // if someone registered for SCO, send to it
                btusb_enqueue_voice(p_dev, p_burst, p_burst[2] + BTUSB_VOICE_HEADER_SIZE);


                p_burst += (p_burst[2] + BTUSB_VOICE_HEADER_SIZE);
                if (((p_burst + BTUSB_VOICE_HEADER_SIZE) < (p_packet + length))
                        && ((p_burst + p_burst[2]) <= (p_packet + length)))
                    continue; // seems like there's at least one more burst

                break;
            }
            else
            {
                BTUSB_ERR("invalid insync data at offset %u: %x %x %x %x %x (pkt %x %x %x %x %x)\n",
                        (unsigned int) (p_burst - p_packet),
                        p_burst[0], p_burst[1],p_burst[2], p_burst[3], p_burst[4],
                        p_packet[0], p_packet[1], p_packet[2], p_packet[3], p_packet[4]);
                p_dev->voice_pending_size = 0;
                p_dev->stats.voicerx_disc_nobodyhdr++;
                p_burst = p_packet; // show no data from here...
                length = 0;
                break;
            }
        }

        // save anything remaining
        if ((p_burst - p_packet + BTUSB_VOICE_HEADER_SIZE) < length)
        {
            // make sure the data still starts from the right header
            if (!btusb_voice_hdr_match(p_dev, p_burst))
            {
                // reset what was in the pending buffer
                BTUSB_ERR("inv tail of %d bytes\n",
                        (int) (length - (p_burst - p_packet)));
                p_dev->voice_pending_size = 0;
                p_dev->stats.voicerx_disc_notailhdr++;
            }
            else
            {
                // length check for those completely paranoid...
                if ((length - (p_burst - p_packet)) < sizeof(p_dev->voice_pending_buffer))
                {
                    memcpy(p_dev->voice_pending_buffer, p_burst, length - (p_burst - p_packet));
                    p_dev->voice_pending_size = length - (int) (p_burst - p_packet);
                }
            }
        }

        // here's what we're going to show user mode
        length = (int) (p_burst - p_packet);
    }
    // are we in sync but not large enough?
    else if (i == 0)
    {
        memcpy(p_dev->voice_pending_buffer, p_packet, length);
        p_dev->voice_pending_size = length;

        // nothing to show user mode
        length = 0;
    }
    else
    {
        unsigned char *p_burst = &p_packet[i];
        unsigned long pending_bytes_sent = 0;
        unsigned char bytes_sent_buffer[VOICE_PENDING_BUFFER_SIZE];

        // note: this "pending buffer" processing assumes that there's
        // only one voice completion routing running at a time.

        // no, we're not in sync.  Need to forward the pending piece
        // and the beginning of this one?
        if ((long) (p_dev->voice_pending_size + i) == (p_dev->voice_pending_buffer[2]
                + BTUSB_VOICE_HEADER_SIZE))
        {
            // go ahead and send one block
            memcpy(&p_dev->voice_pending_buffer[p_dev->voice_pending_size], p_packet, i);
            btusb_enqueue_voice(p_dev, p_dev->voice_pending_buffer, (p_dev->voice_pending_size + i));


            pending_data_forwarded = TRUE;
            pending_bytes_sent = p_dev->voice_pending_size + i;

            // data may be overwritten, so save it for the API sake
            memcpy(bytes_sent_buffer, p_dev->voice_pending_buffer, pending_bytes_sent);
        }
        p_dev->voice_pending_size = 0; // no point to keep anything any more...

        // now forward the middle chunk if any
        while (1)
        {
            if (btusb_voice_hdr_match(p_dev, p_burst))
            {
                btusb_enqueue_voice(p_dev, p_burst, p_burst[2] + BTUSB_VOICE_HEADER_SIZE);


                p_burst += (p_burst[2] + BTUSB_VOICE_HEADER_SIZE);
                if (((p_burst + BTUSB_VOICE_HEADER_SIZE) < (p_packet + length))
                        && ((p_burst + p_burst[2]) <= (p_packet + length)))
                    continue; // seems like there's at least one more burst

                break;
            }
            else
            {
                BTUSB_DBG("inv middle data at offset %u: %x %x %x (pkt %x %x %x)\n",
                        (unsigned int) (p_burst - p_packet), p_burst[0], p_burst[1],
                        p_burst[2], p_packet[0], p_packet[1], p_packet[2]);
                p_dev->stats.voicerx_disc_nobodyhdr2++;
                p_burst = p_packet + i; // show no data from here...
                length = 0;
                break;
            }
        }

        // save anything remaining
        if ((p_burst - p_packet + BTUSB_VOICE_HEADER_SIZE) < length)
        {
            // make sure the data still starts from the right header
            if (!btusb_voice_hdr_match(p_dev, p_burst))
            {
                // reset what was in the pending buffer
                BTUSB_ERR("invalid tail of %d bytes\n",
                        (int) (length - (p_burst - p_packet)));
                p_dev->voice_pending_size = 0;
                p_dev->stats.voicerx_disc_notailhdr++;
            }
            else
            {
                // length check for those completely paranoid...
                if ((unsigned int) (length - (p_burst - p_packet))
                        < sizeof(p_dev->voice_pending_buffer))
                {
                    memcpy(p_dev->voice_pending_buffer, p_burst, length - (p_burst - p_packet));
                    p_dev->voice_pending_size = length - (int) (p_burst - p_packet);
                }
            }
        }

        // here's what we're going to show user mode
        // if there was no pending data delivered
        length = (int) (p_burst - p_packet - i);

        if (pending_data_forwarded)
        {
            // shift the data right for the user mode indication
            // and put the pending data in front
            memmove(&p_packet[pending_bytes_sent], &p_packet[i], length);
            memcpy(p_packet, bytes_sent_buffer, pending_bytes_sent);
            length += pending_bytes_sent;
        }

        // done forwarding voice to the BTKRNL and preparing the user mode buf
    }

    read_voice_active = FALSE;

    // the next read

    btusb_submit_voice_rx(p_dev, p_trans, GFP_ATOMIC);

    // figure out what to do with data.
    if (length) // if really data
    {
        BTUSB_INFO("data lost\n");
    }

unlock:
    read_unlock_irqrestore(&p_dev->xferlock, flags);
}

